// apt.cc
//
//  Copyright 1999,2000,2001 Daniel Burrows
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//
//  Handles basic apt bookeeping that apt-pkg doesn't :)

#include "apt.h"
#include "undo.h"
#include "config_signal.h"

#include <apt-pkg/sourcelist.h>
#include <apt-pkg/pkgcachegen.h>
#include <apt-pkg/error.h>
#include <apt-pkg/configuration.h>
#include <apt-pkg/init.h>
#include <apt-pkg/depcache.h>

#include <assert.h>
#include <stdio.h>
#include <signal.h>
#include <unistd.h>
#include <sys/stat.h>
#include <errno.h>

aptitudeCacheFile *apt_cache_file=NULL;
signalling_config *aptcfg=NULL;
pkgRecords *apt_package_records=NULL;
undo_list *apt_undos=NULL;

string *pendingerr=NULL;
bool erroriswarning=false;

SigC::Signal0<void> cache_closed, cache_reloaded, cache_reload_failed;

void apt_preinit()
{
  signal(SIGPIPE, SIG_IGN);

#ifndef HAVE_LIBAPT_PKG3
  pkgInitialize(*_config);
  // Magic Happens.. :)
#else
  if(pkgInitConfig(*_config))
    pkgInitSystem(*_config, _system);
#endif

  // Allow a user-specific customization file.
  const char *HOME=getenv("HOME");

  if(HOME)
    {
      string cfgloc=HOME;
      cfgloc+="/.aptitude/config";

      // This attempts to test whether the file is available
      if(access(cfgloc.c_str(), R_OK)==0)
	ReadConfigFile(*_config, cfgloc);
    }

  aptcfg=new signalling_config(_config);

  apt_undos=new undo_list;
}

// Hacked from the APT sources
void apt_dumpcfg(const char *root)
{
  const char *HOME=getenv("HOME");

  if(!HOME)
    return;

  string cfgloc=HOME;
  cfgloc+="/.aptitude";

  if(mkdir(cfgloc.c_str(), 0700)<0 && errno!=EEXIST)
    {
      _error->Errno("mkdir", "%s", cfgloc.c_str());
      return;
    }

  cfgloc+="/config";

  FILE *f=fopen(cfgloc.c_str(), "w");

  if(!f)
    {
      _error->Errno("Unable to open %s for writing", cfgloc.c_str());
      return;
    }

  /* Write out all of the configuration directives by walking the 
     configuration tree */
  const Configuration::Item *Top = _config->Tree(root), *Curr=Top;
  if(Curr)
    do
      {
	fprintf(f, "%s \"%s\";\n", Curr->FullTag().c_str(), Curr->Value.c_str());
      
	if (Curr->Child != 0)
	  {
	    Curr = Curr->Child;
	    continue;
	  }
      
	while (Curr != 0 && Curr->Next == 0 && Curr->Parent!=Top)
	  Curr = Curr->Parent;
	if (Curr != 0)
	  Curr = Curr->Next;
      } while(Curr && Curr!=Top);

  fclose(f);
}

// Revert back to the default set of options.  Is it a hack?  Hmmm...
void apt_revertoptions()
{
  _config=new Configuration;

#ifndef HAVE_LIBAPT_PKG3
  pkgInitialize(*_config);
  // Magic Happens.. :)
#else
  // ick?
  if(pkgInitConfig(*_config))
    pkgInitSystem(*_config, _system);
#endif

  delete aptcfg->setcfg(_config);
}

void apt_init(OpProgress *progress_bar, const char *status_fname)
{
  if(!apt_cache_file)
    apt_reload_cache(progress_bar, status_fname);
}

void apt_reload_cache(OpProgress *progress_bar, const char * status_fname)
{
  cache_closed();

  if(apt_package_records)
    {
      delete apt_package_records;
      apt_package_records=NULL;
    }

  if(apt_cache_file)
    {
      delete apt_cache_file;
      apt_cache_file=NULL;
    }

  aptitudeCacheFile *new_file=new aptitudeCacheFile;

  bool open_failed=!new_file->Open(*progress_bar, (getuid()==0), status_fname)
    || _error->PendingError();

  if(open_failed && getuid()==0)
    {
      _error->Discard();

      open_failed=!new_file->Open(*progress_bar, false, status_fname) ||
	_error->PendingError();

      if(!open_failed)
	_error->Warning("Warning: could not lock the cache file.  Opening in read-only mode");
    }

  if(open_failed)
    {
      delete new_file;
      cache_reload_failed();
      return;
    }

  apt_cache_file=new_file;
  apt_package_records=new pkgRecords(*apt_cache_file);

  // Um, good time to clear our undo info.
  apt_undos->clear_items();

  cache_reloaded();
}

pkg_action_state find_pkg_state(pkgCache::PkgIterator pkg)
{
  aptitudeDepCache::StateCache &state=(*apt_cache_file)[pkg];
  aptitudeDepCache::aptitude_state &extstate=(*apt_cache_file)->get_ext_state(pkg);

  if((!pkg.CurrentVer().end() || state.Install()) &&state.InstBroken())
    return state.Delete()?pkg_broken_deleted:pkg_broken;
  else if(state.Delete())
    {
      if(extstate.selection_state!=pkgCache::State::DeInstall &&
	 extstate.selection_state!=pkgCache::State::Purge)
	return pkg_broken_deleted;
      else
	return pkg_remove;
    }
  else if(state.Install())
    {
      if(extstate.selection_state!=pkgCache::State::Install)
	return pkg_added;
      else if(!pkg.CurrentVer().end())
	{
	  if(state.iFlags&pkgDepCache::ReInstall)
	    return pkg_reinstall;
	  else
	    return pkg_upgrade;
	}
      else
	return pkg_install;
    }
  else if(!pkg.VersionList().end() &&
	  !pkg.CurrentVer().end() &&
	  (extstate.selection_state==pkgCache::State::DeInstall ||
	   extstate.selection_state==pkgCache::State::Purge))
    return pkg_broken;

  else if(!pkg.CurrentVer().end() &&
	  state.Status>0 &&
	  state.Keep() &&
	  !state.InstBroken() &&
	  extstate.selection_state==pkgCache::State::Hold)
    return pkg_hold;
  else if(!pkg.VersionList().end() &&
	  extstate.selection_state==pkgCache::State::Install &&
	  state.Status>0 &&
	  state.Keep())
    {
      // If it's upgradable but auto-upgrade is off, assume that it's just held,
      // not broken.
      if(state.Status == 1 && !aptcfg->FindB(PACKAGE "::Auto-Upgrade", true))
	return pkg_hold;
      else
	return pkg_broken;
    }
  else if(state.iFlags&pkgDepCache::ReInstall)
    return pkg_reinstall;

  return pkg_unchanged;
}
