// pkg_grouppolicy.cc
//
//  Copyright 1999,2000,2001 Daniel Burrows
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; see the file COPYING.  If not, write to
//  the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
//  Boston, MA 02111-1307, USA.
//
//  Implementation of stuff that needs implementing.

#include "aptitude.h"

#include "pkg_grouppolicy.h"
#include "pkg_item.h"
#include "pkg_subtree.h"

#include "generic/apt.h"

#include <map>

using namespace std;

// This special tree munges its tag to allow an integer to be prepended to it.
// Ok, it's a dreadful hack.  I admit it.
class pkg_subtree_with_order:public pkg_subtree
{
  string my_tag;
public:
  pkg_subtree_with_order(string name, unsigned char order, bool expand=false)
    :pkg_subtree(name, expand)
  {
    my_tag+=order;
    my_tag+=pkg_subtree::tag();
  }

  virtual const char *tag() const
  {
    return my_tag.c_str();
  }
};

// The following class is a special policy which is used to terminate a
// policy chain:
class pkg_grouppolicy_end:public pkg_grouppolicy
{
public:
  pkg_grouppolicy_end(pkg_signal *_sig)
    :pkg_grouppolicy(_sig) {}

  virtual void add_package(const pkgCache::PkgIterator &i, pkg_subtree *root)
    {
      root->add_child(new pkg_item(i, get_sig()));
    }
};

pkg_grouppolicy *pkg_grouppolicy_end_factory::instantiate(pkg_signal *_sig)
{
  return new pkg_grouppolicy_end(_sig);
}

/*****************************************************************************/

// Groups packages so that the latest version of the package is shown, sorted
// by section.

class pkg_grouppolicy_section:public pkg_grouppolicy
{
  typedef hash_map<string, pair<pkg_grouppolicy *, pkg_subtree *> > section_map;
  section_map sections;

  pkg_grouppolicy_factory *chain;

  // As in the factory
  int split_mode;
  bool passthrough;

  static hash_map<string, const char *> section_descriptions;
  static void init_section_descriptions();
public:
  pkg_grouppolicy_section(int _split_mode,
			  bool _passthrough,
			  pkg_grouppolicy_factory *_chain,
			  pkg_signal *_sig)
    :pkg_grouppolicy(_sig), chain(_chain),
     split_mode(_split_mode), passthrough(_passthrough)
  {
    init_section_descriptions();
  }

  virtual void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root);

  virtual ~pkg_grouppolicy_section()
    {
      for(section_map::iterator i=sections.begin(); i!=sections.end(); i++)
        delete i->second.first;
    }
};

pkg_grouppolicy *pkg_grouppolicy_section_factory::instantiate(pkg_signal *_sig)
{
  return new pkg_grouppolicy_section(split_mode, passthrough, chain, _sig);
}

hash_map<string, const char *> pkg_grouppolicy_section::section_descriptions;

// Should this be externally configurable?
void pkg_grouppolicy_section::init_section_descriptions()
{
  static bool already_done=false;

  if(already_done)
    return;

  section_descriptions["Tasks"]=_("Packages which set up your computer to perform a particular task");
  section_descriptions["Unknown"]=_("Packages with no declared section");
  section_descriptions["admin"]=_("Adminitrative utilities (install software, manage users, etc)");
  section_descriptions["base"]=_("The Debian base system");
  section_descriptions["comm"]=_("Programs which allow you to use a faxmodem");
  section_descriptions["devel"]=_("Utilities and programs for software development");
  section_descriptions["doc"]=_("Documentation and specialized programs for viewing documentation");
  section_descriptions["editors"]=_("Text editors and word processors");
  section_descriptions["electronics"]=_("Programs for working with circuits and electronics");
  section_descriptions["games"]=_("Games, toys, and fun programs");
  section_descriptions["graphics"]=_("Utilities to create, view, and edit graphics files");
  section_descriptions["hamradio"]=_("Software for ham radio operators");
  section_descriptions["interpreters"]=_("Interpreters for interpreted languages");
  section_descriptions["libs"]=_("Collections of software routines");
  section_descriptions["mail"]=_("Programs to write, send, and route email messages");
  section_descriptions["math"]=_("Numeric analysis and other mathematics-related software");
  section_descriptions["misc"]=_("Miscellaneous software");
  section_descriptions["net"]=_("Programs to connect to and provide various services");
  section_descriptions["news"]=_("Usenet clients and servers");
  section_descriptions["oldlibs"]=_("Obsolete libraries");
  section_descriptions["otherosfs"]=_("Emulators and software to read foreign filesystems");
  section_descriptions["science"]=_("Software for scientific work");
  section_descriptions["shells"]=_("Command shells and alternative console environments");
  section_descriptions["sound"]=_("Utilities to play and record sound");
  section_descriptions["tex"]=_("The TeX typesetting system");
  section_descriptions["text"]=_("Text processing utilities");
  section_descriptions["utils"]=_("Various system utilities");
  section_descriptions["web"]=_("Web browsers, servers, proxies, and other tools");
  section_descriptions["x11"]=_("The X window system and related software");

  section_descriptions["contrib"]=_("Programs which depend on software not in Debian");
  section_descriptions["main"]=_("The main Debian archive");
  section_descriptions["non-US"]=_("Programs stored outside the US due to export controls");
  section_descriptions["non-free"]=_("Programs which are not free software");

  already_done=true;
}

void pkg_grouppolicy_section::add_package(const pkgCache::PkgIterator &pkg,
					  pkg_subtree *root)
{
  const char *section=pkg.VersionList().Section();
  const char *name=pkg.Name();
  bool maypassthrough=false; // FIXME: HACK!

  if(name[0] == 't' && name[1] == 'a' && name[2] == 's' && name[3] == 'k' && name[4] == '-')
    {
      maypassthrough=true;
      section=split_mode!=pkg_grouppolicy_section_factory::split_none?_("Tasks/Tasks"):_("Tasks");
    }

  if(!section)
    {
      maypassthrough=true;
      section=split_mode!=pkg_grouppolicy_section_factory::split_none?_("Unknown/Unknown"):_("Unknown");
    }

  if(!*section)
    {
      maypassthrough=true;
      section=split_mode!=pkg_grouppolicy_section_factory::split_none?_("virtual/virtual"):_("virtual");
    }

  const char *split=strchr(section, '/');
  const char *subdir=split?split+1:section;

  string tag;

  if(split_mode==pkg_grouppolicy_section_factory::split_none)
    tag=section;
  else if(split_mode==pkg_grouppolicy_section_factory::split_topdir)
    tag=(split?string(section,split-section):string(_("main")));
  else if(split_mode==pkg_grouppolicy_section_factory::split_subdir)
    tag=subdir;

  section_map::iterator found=sections.find(tag);

  if(maypassthrough && passthrough)
    {
      if(found==sections.end())
	sections[tag].first=chain->instantiate(get_sig());

      sections[tag].first->add_package(pkg, root);
    }
  else
    {
      if(found==sections.end())
	{
	  pkg_subtree *newtree;
	  if(section_descriptions.find(tag)!=section_descriptions.end())
	    newtree=new pkg_subtree(tag+": "+section_descriptions[tag]);
	  else
	    newtree=new pkg_subtree(tag);

	  sections[tag].first=chain->instantiate(get_sig());
	  sections[tag].second=newtree;

	  root->add_child(newtree);
	}

      sections[tag].first->add_package(pkg, sections[tag].second);
    }
}

/*****************************************************************************/

class pkg_grouppolicy_status:public pkg_grouppolicy
{
  static const int numgroups=6;
  pkg_grouppolicy_factory *chain;

  enum states {newpkg, upgradable, installed, not_installed, obsolete_pkg, virtual_pkg};

  static const char * const state_titles[numgroups];
  // FIXME: need better titles :)

  pair<pkg_grouppolicy *, pkg_subtree *> children[numgroups];
public:
  pkg_grouppolicy_status(pkg_grouppolicy_factory *_chain, pkg_signal *_sig)
    :pkg_grouppolicy(_sig), chain(_chain)
    {
      for(int i=0; i<( (int) (sizeof(children)/sizeof(children[0]))); i++)
	{
	  children[i].first=NULL;
	  children[i].second=NULL;
	}
    }

  virtual void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root);

  virtual ~pkg_grouppolicy_status()
    {
      for(int i=0; i<4; i++)
	delete children[i].first;
    }
};

const char * const pkg_grouppolicy_status::state_titles[numgroups] =
{
  N_("New Packages"),
  N_("Upgradable Packages"),
  N_("Installed Packages"),
  N_("Not Installed Packages"),
  N_("Obsolete and Locally Created Packages"),
  N_("Virtual Packages")
};

pkg_grouppolicy *pkg_grouppolicy_status_factory::instantiate(pkg_signal *_sig)
{
  return new pkg_grouppolicy_status(chain, _sig);
}

void pkg_grouppolicy_status::add_package(const pkgCache::PkgIterator &pkg,
					 pkg_subtree *root)
{
  states section;
  pkgDepCache::StateCache &state=(*apt_cache_file)[pkg];

  if(pkg.VersionList().end())
    section=virtual_pkg;
  else
    // Yuck.  Obsolete packages can be detected because they have only a single
    // file entry.  Of course, libapt makes it impossible to easily check this.
    {
      bool obsolete=true;

      if(pkg.CurrentVer().end())
	obsolete=false;
      else
	{
	  pkgCache::VerIterator ver=pkg.VersionList();
	  ver++;

	  if(!ver.end())
	    obsolete=false;
	  else // Ok, there's only one version.  Good.
	    {
	      pkgCache::VerFileIterator files=pkg.VersionList().FileList();
	      if(!files.end())
		{
		  files++;
		  if(!files.end())
		    obsolete=false; // Nope, more than one file
		}
	    }
	}
      if(obsolete)
	section=obsolete_pkg;
      else if((*apt_cache_file)->get_ext_state(pkg).new_package)
	section=newpkg;
      else if(state.Status==2)
	section=not_installed;
      else if(state.Upgradable())
	section=upgradable;
      else
	section=installed;
    }

  if(!children[section].second)
    {
      pkg_subtree *newtree=new pkg_subtree_with_order(_(state_titles[section]),
						      section);
      children[section].first=chain->instantiate(get_sig());
      children[section].second=newtree;
      root->add_child(newtree);
    }

  children[section].first->add_package(pkg, children[section].second);
}

/*****************************************************************************/
class pkg_grouppolicy_filter:public pkg_grouppolicy
{
  pkg_filter filter;

  pkg_grouppolicy *chain;
public:
  pkg_grouppolicy_filter(pkg_grouppolicy_factory *_chain, pkg_filter _filter,
			 pkg_signal *_sig)
    :pkg_grouppolicy(_sig),
     filter(_filter),
     chain(_chain->instantiate(_sig))
  {
  }

  virtual void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root)
  {if(filter(pkg)) chain->add_package(pkg, root);}

  virtual ~pkg_grouppolicy_filter() {delete chain;}
};

pkg_grouppolicy *pkg_grouppolicy_filter_factory::instantiate(pkg_signal *_sig)
{
  return new pkg_grouppolicy_filter(chain, filter, _sig);
}

bool pkg_missing_filter(const pkgCache::PkgIterator &pkg)
{
  return (!pkg.VersionList().end()) || (!pkg.ProvidesList().end());
}

/*****************************************************************************/

class pkg_grouppolicy_mode:public pkg_grouppolicy
{
  const static char * const child_names[];
  pair<pkg_grouppolicy *, pkg_subtree *> children[num_pkg_action_states];
  pkg_grouppolicy_factory *chain;
public:
  pkg_grouppolicy_mode(pkg_grouppolicy_factory *_chain, pkg_signal *_sig)
    :pkg_grouppolicy(_sig), chain(_chain)
  {
    for(int i=0; i<num_pkg_action_states; i++)
      {
	children[i].first=NULL;
	children[i].second=NULL;
      }
  }

  void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root)
  {
    int group=find_pkg_state(pkg);
    if(group!=pkg_unchanged)
      {
	if(!children[group].second)
	  {
	    pkg_subtree *newtree=new pkg_subtree_with_order(_(child_names[group]),
							    group,
							    true);
	    root->add_child(newtree);
	    children[group].first=chain->instantiate(get_sig());
	    children[group].second=newtree;
	  }

	children[group].first->add_package(pkg, children[group].second);
      }
  }

  virtual ~pkg_grouppolicy_mode()
  {
    for(int i=0; i<num_pkg_action_states; i++)
      delete children[i].first;
  }
};

const char * const pkg_grouppolicy_mode::child_names[]=
{
  N_("Packages to be installed"),
  N_("Packages to be upgraded"),
  N_("Packages to be removed"),
  N_("Packages to be reinstalled"),
  N_("Packages being held back"),
  N_("Packages being held in their current state due to broken dependencies"),
  N_("Packages being deleted due to broken dependencies"),
  N_("Packages being installed to satisfy dependencies")
};

pkg_grouppolicy *pkg_grouppolicy_mode_factory::instantiate(pkg_signal *_sig)
{
  return new pkg_grouppolicy_mode(chain, _sig);
}

/*****************************************************************************/

class pkg_grouppolicy_firstchar:public pkg_grouppolicy
{
  pkg_grouppolicy_factory *chain;

  typedef map<char, pair<pkg_grouppolicy *, pkg_subtree *> > childmap;
  // Store the child group policies and their associated subtrees.
  childmap children;
public:
  pkg_grouppolicy_firstchar(pkg_grouppolicy_factory *_chain, pkg_signal *_sig)
    :pkg_grouppolicy(_sig), chain(_chain)
  {
  }

  ~pkg_grouppolicy_firstchar()
  {
    for(childmap::iterator i=children.begin(); i!=children.end(); i++)
      delete i->second.first;
  }

  void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root)
  {
    assert(pkg.Name());

    char firstchar=toupper(pkg.Name()[0]);

    childmap::iterator found=children.find(firstchar);
    if(found!=children.end())
      found->second.first->add_package(pkg, found->second.second);
    else
      {
	string treename;
	treename+=firstchar;

	pkg_subtree *newtree=new pkg_subtree(treename);
	pkg_grouppolicy *newchild=chain->instantiate(get_sig());
	children[firstchar].first=newchild;
	children[firstchar].second=newtree;
	root->add_child(newtree);

	newchild->add_package(pkg, newtree);
      }
  }
};

pkg_grouppolicy *pkg_grouppolicy_firstchar_factory::instantiate(pkg_signal *sig)
{
  return new pkg_grouppolicy_firstchar(chain, sig);
}

/*****************************************************************************/

// Groups packages by priority
class pkg_grouppolicy_priority:public pkg_grouppolicy
{
  // a map may be overkill, but I figure better safe than sorry..
  // who knows, maybe someone will change the number of priorities someday..
  typedef map<unsigned char,
	      pair<pkg_grouppolicy *, pkg_subtree *> > childmap;

  childmap children;
  pkg_grouppolicy_factory *chain;

  pkg_grouppolicy *spillover;
public:
  pkg_grouppolicy_priority(pkg_grouppolicy_factory *_chain, pkg_signal *_sig)
    :pkg_grouppolicy(_sig),
     chain(_chain),
     spillover(_chain->instantiate(get_sig()))
  {
  }

  ~pkg_grouppolicy_priority()
  {
    for(childmap::iterator i=children.begin(); i!=children.end(); i++)
      delete i->second.first;
  }

  void add_package(const pkgCache::PkgIterator &pkg, pkg_subtree *root)
  {
    unsigned char priority;
    const char *pstr;

    if(!pkg.VersionList().end())
      {
	if(!pkg.CurrentVer().end())
	  priority=pkg.CurrentVer()->Priority;
	else
	  priority=pkg.VersionList()->Priority;

#ifdef HAVE_LIBAPT_PKG3
	pstr=(*apt_cache_file)->GetCache().Priority(priority);
#else
	pstr=(*apt_cache_file)->Priority(priority);
#endif
      }
    else
      {
	pstr=NULL;
	priority=0;
      }

    // Some packages don't have a reasonable priority
    if(!pstr)
      {
	pstr=_("unknown");
	priority=0;
      }

    childmap::iterator found=children.find(priority);

    if(found!=children.end())
      found->second.first->add_package(pkg, found->second.second);
    else
      {
	char buf[512];
	snprintf(buf, 512, _("Priority %s"), pstr);
	int order;
	switch(priority)
	  {
	  case pkgCache::State::Required : order=1; break;
	  case pkgCache::State::Important: order=2; break;
	  case pkgCache::State::Standard : order=3; break;
	  case pkgCache::State::Optional : order=4; break;
	  case pkgCache::State::Extra    : order=5; break;
	  default                        : order=6; break;
	  };

	pkg_subtree *newtree=new pkg_subtree_with_order(buf, order);
	pkg_grouppolicy *newchild=chain->instantiate(get_sig());
	children[priority].first=newchild;
	children[priority].second=newtree;
	root->add_child(newtree);

	newchild->add_package(pkg, newtree);
      }
  }  
};

pkg_grouppolicy *pkg_grouppolicy_priority_factory::instantiate(pkg_signal *sig)
{
  return new pkg_grouppolicy_priority(chain, sig);
}

