// vs_editline.cc
//
//  Copyright 2000 Daniel Burrows

#include "vs_editline.h"
#include "config/keybindings.h"
#include "config/colors.h"
#include "vscreen.h"

using namespace std;

keybindings *vs_editline::bindings=NULL;

vs_editline::vs_editline(string _prompt, string _text,
			 history_list *_history)
  :vscreen_widget(),prompt(_prompt), text(_text), curloc(_text.size()),
   desired_size(-1), history(_history), history_loc(0), using_history(false)
{
  set_bg(get_color("EditLine"));

  // This ensures that the cursor is set to the right location when the
  // widget is displayed or resized:
  do_layout.connect(SigC::slot(vscreen_updatecursor));
}

vs_editline::vs_editline(int maxlength, const string &_prompt,
			 const string &_text, history_list *_history)
  :vscreen_widget(),prompt(_prompt), text(_text), curloc(_text.size()),
   desired_size(maxlength), history(_history), history_loc(0),
   using_history(true)
{
  set_bg(get_color("EditLine"));
  do_layout.connect(SigC::slot(vscreen_updatecursor));
}

bool vs_editline::focus_me()
{
  return true;
}

bool vs_editline::handle_char(chtype ch)
{
  if(bindings->key_matches(ch, "DelBack"))
    {
      if(curloc>0)
	{
	  text.erase(--curloc, 1);
	  vscreen_update();
	}
      else
	{
	  beep();
	  //	  refresh();
	}
      return true;
    }
  else if(bindings->key_matches(ch, "DelForward"))
    {
      if(curloc<text.size())
	{
	  text.erase(curloc, 1);
	  vscreen_update();
	}
      else
	{
	  beep();
	  //	  refresh();
	}
      return true;
    }
  else if(bindings->key_matches(ch, "Confirm"))
    {
      entered(text);
      return true;
    }
  else if(bindings->key_matches(ch, "Left"))
    {
      if(curloc>0)
	{
	  curloc--;
	  vscreen_update();
	}
      else
	{
	  beep();
	  //	  refresh();
	}
      return true;
    }
  else if(bindings->key_matches(ch, "Right"))
    {
      if(curloc<text.size())
	{
	  curloc++;
	  vscreen_update();
	}
      else
	{
	  beep();
	  //	  refresh();
	}
      return true;
    }
  else if(bindings->key_matches(ch, "Begin"))
    {
      curloc=0;
      vscreen_update();
      return true;
    }
  else if(bindings->key_matches(ch, "End"))
    {
      curloc=text.size();
      vscreen_update();
      return true;
    }
  else if(bindings->key_matches(ch, "DelEOL"))
    {
      text.erase(curloc);
      vscreen_update();
      return true;
    }
  else if(bindings->key_matches(ch, "DelBOL"))
    {
      text.erase(0, curloc);
      curloc=0;
      vscreen_update();
      return true;
    }
  else if(history && bindings->key_matches(ch, "HistoryPrev"))
    {
      if(history->size()==0)
	return true;

      if(!using_history)
	{
	  using_history=true;
	  history_loc=history->size()-1;
	  pre_history_text=text;
	}
      else if(history_loc>0)
	--history_loc;
      else
	// Break out
	return true;

      text=(*history)[history_loc];
      curloc=text.size();
      vscreen_update();

      return true;
    }
  else if(history && bindings->key_matches(ch, "HistoryNext"))
    {
      if(history->size()==0 || !using_history)
	return true;

      if(history_loc>=history->size()-1)
	{
	  using_history=false;
	  history_loc=0;
	  text=pre_history_text;
	  pre_history_text="";
	  curloc=text.size();
	  vscreen_update();

	  // FIXME: store the pre-history edit and restore that.
	  return true;
	}

      if(history_loc>=0)
	{
	  ++history_loc;
	  text=(*history)[history_loc];
	  curloc=text.size();
	  vscreen_update();

	  return true;
	}
      else
	return true;
    }
  else if(ch>255)
    return vscreen_widget::handle_char(ch);
  else if(ch=='\t') // HACK
    return false;
  else
    {
      text.insert(curloc++, 1, (char) ch);
      text_changed();
      vscreen_update();
      return true;
    }
}

void vs_editline::paint()
{
  int width=getmaxx();

  int dispoff=(prompt.size()+curloc)-(prompt.size()+curloc)%width;
  string todisp(prompt+text, dispoff, width);

  mvaddstr(0, 0, todisp.c_str());
}

void vs_editline::dispatch_mouse(short id, int x, int y, int z, mmask_t bstate)
{
  int dispoff=(prompt.size()+curloc)-(prompt.size()+curloc)%getmaxx();

  string::size_type mouseloc=x+dispoff-prompt.size();

  if(mouseloc>=0 && mouseloc<=text.size())
    curloc=mouseloc;
  else if(mouseloc>text.size())
    curloc=text.size();
  else
    return; // Break out

  vscreen_update();
}

void vs_editline::add_to_history(std::string s,
				 history_list *lst)
{
  assert(lst);

  if(lst->empty() || lst->back()!=s)
    lst->push_back(s);
}

void vs_editline::add_to_history(std::string s)
{
  if(history)
    add_to_history(s, history);
}

void vs_editline::reset_history()
{
  pre_history_text="";
  using_history=false;
  history_loc=0;
}

void vs_editline::set_text(string _text)
{
  text=_text;
  if(curloc>text.size())
    curloc=text.size();
  text_changed();
  vscreen_update();
}

void vs_editline::init_bindings()
{
  bindings=new keybindings(&global_bindings);

  bindings->set("Left", KEY_LEFT);
  bindings->set("Right", KEY_RIGHT);
  // Override these for the case where left and right have multiple bindings
  // in the global keymap
}
